//=============================================================================
//
//  CLASS Triangle - IMPLEMENTATION
//
//=============================================================================

#define RAYTRACER_TRIANGLE_CPP


//== INCLUDES =================================================================
#include "TriangleSoup.h"
#include "Triangle.h"
#include "Tools.h"

   
//== IMPLEMENTATION ========================================================== 

Triangle::Triangle( TriangleSoup* pclMesh,
		       int   i3VertexIndex[3],
			   int   i3NormalIndex[3],
			   int   faceIndex ) 
{
    m_faceIndex = faceIndex;
	m_pclMesh = pclMesh;
	if ( !m_pclMesh ) std::cerr << "invalid mesh-reference";

	memcpy( m_i3VertexIndex, i3VertexIndex, sizeof(m_i3VertexIndex) );
	memcpy( m_i3NormalIndex, i3NormalIndex, sizeof(m_i3NormalIndex) );
}
	
//-----------------------------------------------------------------------------
	
AABBox 
Triangle::getAABB() const
{
	AABBox cl_aabbox;
	
	cl_aabbox.extendBy( m_pclMesh->getVertexList()[m_i3VertexIndex[0]]);
	cl_aabbox.extendBy( m_pclMesh->getVertexList()[m_i3VertexIndex[1]]);
	cl_aabbox.extendBy( m_pclMesh->getVertexList()[m_i3VertexIndex[2]]);
	return cl_aabbox;
}

//-----------------------------------------------------------------------------
//Schnittfunktion zwischen Strahl und Dreieck
bool 
Triangle::intersect(const Ray& clRay, double t0, double t1, TracingContext& tContext ) const
{	
	//Die 3 Punkte des Dreiecks
	VectorType3 v3_point0 = m_pclMesh->getVertexList()[m_i3VertexIndex[0]]; 
	VectorType3 v3_point1 = m_pclMesh->getVertexList()[m_i3VertexIndex[1]];
	VectorType3 v3_point2 = m_pclMesh->getVertexList()[m_i3VertexIndex[2]];

	VectorType3 u = v3_point1-v3_point0;
	VectorType3 v = v3_point2-v3_point0;
	
	//Der Normalenvektor der durch das Dreieck aufgespannten Ebene
	VectorType3 n = u % v;
	n.normalize();
	
    /*
		Strahl-Dreieck-Schnittest.
	*/

	// get intersect point of ray with triangle plane
	double r_denominator = n | clRay.getDir();
		
	if ( isZero(r_denominator) ) return false;
	double t = ( n|(v3_point0 - clRay.getOrigin()) ) / r_denominator;

	// reject, if not in interval
	if ( t < t0 || t > t1 )  return false;

	VectorType3 v3_intersection = clRay.evaluate( t );		// intersect point of ray and plane

	// is I inside T?
	double uu = u | u;
	double uv = u | v ;
	double vv = v | v;

	VectorType3 w = v3_intersection - v3_point0;
	double wu = w | u;
	double wv = w | v;
	double D = uv * uv - uu * vv;		// this is the double area between v and u squared
	
	// get and test barycentric coords
	double beta = (uv * wv - vv * wu) / D;		// the weight for u and thus vertex 2
	if (beta < -0.001 || beta > 1.001)				// I is outside T
		return false;
	double gamma = (uv * wu - uu * wv) / D;		// the weight for v and thus vertex 3
	if (gamma < -0.001 || (beta + gamma) > 1.001)		// I is outside T
		return false;
	
	//Setzen der Parameter fr 
	tContext.m_t = t;
	tContext.m_v3HitPoint = v3_intersection;
	//Nicht in allen OBJ-Dateien werden die Vertex-Normalen auch gesetzt
	//Ist dies nicht der Fall, verwenden wir einfach die Ebenen-Normale
	if ( m_i3NormalIndex[0] != -1 &&
	     m_i3NormalIndex[0] != -1 &&
	     m_i3NormalIndex[0] != -1 )
	    tContext.m_v3Normal = (1-beta-gamma)*m_pclMesh->getNormalList()[m_i3NormalIndex[0]]+
		                              beta*m_pclMesh->getNormalList()[m_i3NormalIndex[1]]+
			    		             gamma*m_pclMesh->getNormalList()[m_i3NormalIndex[2]];
	else
	    tContext.m_v3Normal = n;
	tContext.m_v3Normal.normalize();

	tContext.m_pclShader = m_pclMesh->getShader(m_faceIndex);
	return true;				
}

//Render in GL Mode
void Triangle::renderGL() const
{
    if ( !m_pclMesh->drawnThisCycle() )
        m_pclMesh->renderGL();
}



int Triangle::getNormalIndex( unsigned int i ) const
{
    if ( i < 3 )
        return m_i3NormalIndex[ i ];
    return -1;
}

int Triangle::getVertexIndex( unsigned int i ) const
{
    if ( i < 3 )
        return m_i3VertexIndex[ i ];
    return -1;
}


void Triangle::setSelected( bool sel )   
{
    m_pclMesh->setSelected( sel );
};


bool Triangle::getSelected() const
{
    return m_pclMesh->getSelected();
};

void    Triangle::translate( float x, float y, float z )
{
    m_pclMesh->translateRelative( VectorType3( x, y, z ) );
};